using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;

namespace RenderToCube
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;

        RenderTargetCube renderCube;
        DepthStencilBuffer depthStencil;

        KeyboardState lastKbState;

        Model sphere;
        Model walls;
        Model columns;

        Effect[] lightEffect = new Effect[3];
        int currentEffect = 0;
        Effect envMapping;

        Vector3 cameraPos = new Vector3(0, 2, 4.8f);
        Vector3 lightPosition;
        Vector3 lightDirection;

        Matrix view, projection;

        Vector3 spherePos = Vector3.Zero;

        Camera camera;

        KeyboardState lastKbdState = Keyboard.GetState();
        MouseState lastMsState = Mouse.GetState();

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            // TODO: Add your initialization logic here

            base.Initialize();
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);

            lightEffect[0] = Content.Load<Effect>("PointLight");
            lightEffect[1] = Content.Load<Effect>("DirectionalLight");
            lightEffect[2] = Content.Load<Effect>("SpotLight");
            envMapping = Content.Load<Effect>("EnvMapping");

            sphere = Content.Load<Model>("SphereHighPoly");
            walls = Content.Load<Model>("BathHouse-Walls");
            columns = Content.Load<Model>("BathHouse-Cols");
            //walls = Content.Load<Model>("test_scene");


            camera = new Camera(cameraPos, -Vector3.UnitZ);
            view = Matrix.CreateLookAt(cameraPos, cameraPos - Vector3.UnitZ, Vector3.Up);
            float aspectRatio = (float)graphics.GraphicsDevice.Viewport.Width / (float)graphics.GraphicsDevice.Viewport.Height;
            projection = Matrix.CreatePerspectiveFieldOfView(MathHelper.ToRadians(75), aspectRatio, 0.1f, 100);

            lightEffect[0].Parameters["View"].SetValue(view);
            lightEffect[0].Parameters["Projection"].SetValue(projection);

            lightEffect[0].Parameters["cameraPosition"].SetValue(cameraPos);
            lightEffect[0].Parameters["specularPower"].SetValue(50);
            lightEffect[0].Parameters["specularIntensity"].SetValue(0);

            lightPosition = new Vector3(0, 2.5f, 0);
            lightDirection = new Vector3(-1.0f, -0.5f, -0.1f);
            lightDirection.Normalize();
            //---
            lightEffect[0].Parameters["lightPosition"].SetValue(lightPosition);
            SetLightColors(lightEffect[0]);


            //--            
            lightEffect[1].Parameters["lightDirection"].SetValue(lightDirection);
            SetLightColors(lightEffect[1]);

            //--
            lightEffect[2].Parameters["lightPosition"].SetValue(lightPosition);
            lightEffect[2].Parameters["lightDirection"].SetValue(lightDirection);
            lightEffect[2].Parameters["cosInnerCone"].SetValue((float)Math.Cos(MathHelper.ToRadians(45)));
            lightEffect[2].Parameters["cosOutterCone"].SetValue((float)Math.Cos(MathHelper.ToRadians(60)));
            SetLightColors(lightEffect[2]);

            //---
            envMapping.Parameters["lightPosition"].SetValue(lightPosition);
            SetLightColors(envMapping);

            //--
            ChangeEffect(walls, lightEffect[currentEffect]);
            ChangeEffect(columns, lightEffect[currentEffect]);

            ChangeEffect(sphere, envMapping);

            //--
            CreateRenderTarget();
        }

        private void SetLightColors(Effect effect)
        {
            effect.Parameters["ambientLightColor"].SetValue(new Vector4(0.1f, 0.1f, 0.1f, 1));
            effect.Parameters["diffuseLightColor"].SetValue(new Vector4(0.5f, 0.5f, 0.5f, 1));
            effect.Parameters["specularLightColor"].SetValue(new Vector4(1.0f, 1.0f, 1.0f, 1));
        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            KeyboardState kbState = Keyboard.GetState();
            MouseState msState = Mouse.GetState();

            // Allows the game to exit
            if (kbState.IsKeyDown(Keys.Escape))
                this.Exit();

            if (kbState.IsKeyDown(Keys.Space) && !lastKbState.IsKeyDown(Keys.Space))
            {
                currentEffect = (currentEffect + 1) % lightEffect.Length;
                ChangeEffect(walls, lightEffect[currentEffect]);
                ChangeEffect(columns, lightEffect[currentEffect]);
            }

            if (kbState.IsKeyDown(Keys.Right))
            {
                lightDirection = Vector3.TransformNormal(lightDirection, Matrix.CreateRotationY(-0.1f));
                lightEffect[1].Parameters["lightDirection"].SetValue(lightDirection);
                lightEffect[2].Parameters["lightDirection"].SetValue(lightDirection);
            }
            else if (kbState.IsKeyDown(Keys.Left))
            {
                lightDirection = Vector3.TransformNormal(lightDirection, Matrix.CreateRotationY(0.1f));
                lightEffect[1].Parameters["lightDirection"].SetValue(lightDirection);
                lightEffect[2].Parameters["lightDirection"].SetValue(lightDirection);
            }

            float delta = (float)gameTime.ElapsedGameTime.TotalSeconds;
            MoveSphere(kbState, delta);

            // Camera Handling
            UpdateCamera(ref kbState, ref msState, delta);

            lastKbState = kbState;
            lastMsState = msState;

            base.Update(gameTime);
        }

        private void UpdateCamera(ref KeyboardState kbdState, ref MouseState msState, float frameSeconds)
        {
            float f = 0, r = 0;
            if (kbdState.IsKeyDown(Keys.W))
                f = frameSeconds;
            else if (kbdState.IsKeyDown(Keys.S))
                f = -frameSeconds;
            if (kbdState.IsKeyDown(Keys.D))
                r = frameSeconds;
            else if (kbdState.IsKeyDown(Keys.A))
                r = -frameSeconds;

            float rotUpDown = 0;
            float rotRightLeft = 0;

            if (lastMsState.RightButton == ButtonState.Pressed || msState.RightButton == ButtonState.Pressed)
            {
                rotUpDown = lastMsState.Y - msState.Y;
                rotRightLeft = lastMsState.X - msState.X;
            }

            if (f != 0 || r != 0 || rotUpDown != 0 || rotRightLeft != 0)
            {
                camera.Move(f, r);
                camera.Rotate(rotUpDown, rotRightLeft);
                view = camera.View;
                cameraPos = camera.Position;
                envMapping.Parameters["cameraPosition"].SetValue(cameraPos);
            }
        }

        private KeyboardState MoveSphere(KeyboardState kbState, float delta)
        {
            if (kbState.IsKeyDown(Keys.NumPad4))
            {
                spherePos.X -= delta;
            }
            else if (kbState.IsKeyDown(Keys.NumPad6))
            {
                spherePos.X += delta;
            }
            if (kbState.IsKeyDown(Keys.NumPad8))
            {
                spherePos.Z -= delta;
            }
            else if (kbState.IsKeyDown(Keys.NumPad2))
            {
                spherePos.Z += delta;
            }
            if (kbState.IsKeyDown(Keys.NumPad7))
            {
                spherePos.Y -= delta;
            }
            else if (kbState.IsKeyDown(Keys.NumPad9))
            {
                spherePos.Y += delta;
            }
            return kbState;
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            RenderToCube(spherePos);

            GraphicsDevice.Clear(Color.CornflowerBlue);

            DrawModel(walls, Matrix.Identity);
            DrawModel(columns, Matrix.Identity);
            DrawModel(sphere, Matrix.CreateTranslation(spherePos));

            base.Draw(gameTime);
        }

        private void DrawModel(Model model, Matrix world)
        {
            Matrix[] transforms = new Matrix[model.Bones.Count];
            model.CopyAbsoluteBoneTransformsTo(transforms);

            foreach (ModelMesh mesh in model.Meshes)
            {
                foreach (Effect effect in mesh.Effects)
                {
                    effect.Parameters["World"].SetValue(transforms[mesh.ParentBone.Index] * world);
                }
                mesh.Draw();
            }
        }

        public void ChangeEffect(Model model, Effect effect)
        {
            foreach (ModelMesh mesh in model.Meshes)
            {
                foreach (ModelMeshPart part in mesh.MeshParts)
                {
                    part.Effect = effect;
                }
            }
        }

        private void CreateRenderTarget()
        {
            renderCube = new RenderTargetCube(GraphicsDevice, 512, 1, SurfaceFormat.Rgba64);
            depthStencil = new DepthStencilBuffer(GraphicsDevice, 512, 512,
        GraphicsDevice.DepthStencilBuffer.Format);
        }

        private void RenderToCube(Vector3 center)
        {
            DepthStencilBuffer oldDS = GraphicsDevice.DepthStencilBuffer;
            Matrix rtcView;
            Matrix rtcProjection;

            GraphicsDevice.DepthStencilBuffer = depthStencil;
            GraphicsDevice.RenderState.CullMode = CullMode.CullClockwiseFace;

            rtcProjection = Matrix.CreatePerspectiveFieldOfView(MathHelper.PiOver2, 1, 0.1f, 100);
            lightEffect[0].Parameters["Projection"].SetValue(rtcProjection);

            rtcView = Matrix.CreateLookAt(center, center + Vector3.UnitX, Vector3.UnitY) * Matrix.CreateScale(-1, 1, 1);
            lightEffect[0].Parameters["View"].SetValue(rtcView);
            GraphicsDevice.SetRenderTarget(0, renderCube, CubeMapFace.PositiveX);
            GraphicsDevice.Clear(Color.Red);
            DrawModel(walls, Matrix.Identity);
            DrawModel(columns, Matrix.Identity);


            rtcView = Matrix.CreateLookAt(center, center - Vector3.UnitX, Vector3.UnitY) * Matrix.CreateScale(-1, 1, 1);
            lightEffect[0].Parameters["View"].SetValue(rtcView);
            GraphicsDevice.SetRenderTarget(0, renderCube, CubeMapFace.NegativeX);
            GraphicsDevice.Clear(Color.Red);
            DrawModel(walls, Matrix.Identity);
            DrawModel(columns, Matrix.Identity);


            rtcView = Matrix.CreateLookAt(center, center + Vector3.UnitZ, Vector3.UnitY) * Matrix.CreateScale(-1, 1, 1);
            lightEffect[0].Parameters["View"].SetValue(rtcView);
            GraphicsDevice.SetRenderTarget(0, renderCube, CubeMapFace.PositiveZ);
            GraphicsDevice.Clear(Color.Blue);
            DrawModel(walls, Matrix.Identity);
            DrawModel(columns, Matrix.Identity);


            rtcView = Matrix.CreateLookAt(center, center - Vector3.UnitZ, Vector3.UnitY) * Matrix.CreateScale(-1, 1, 1);
            lightEffect[0].Parameters["View"].SetValue(rtcView);
            GraphicsDevice.SetRenderTarget(0, renderCube, CubeMapFace.NegativeZ);
            GraphicsDevice.Clear(Color.Blue);
            DrawModel(walls, Matrix.Identity);
            DrawModel(columns, Matrix.Identity);

            rtcView = Matrix.CreateLookAt(center, center + Vector3.UnitY, -Vector3.UnitZ) * Matrix.CreateScale(-1, 1, 1);
            lightEffect[0].Parameters["View"].SetValue(rtcView);
            GraphicsDevice.SetRenderTarget(0, renderCube, CubeMapFace.PositiveY);
            GraphicsDevice.Clear(Color.Green);
            DrawModel(walls, Matrix.Identity);
            DrawModel(columns, Matrix.Identity);

            rtcView = Matrix.CreateLookAt(center, center - Vector3.UnitY, Vector3.UnitZ) * Matrix.CreateScale(-1, 1, 1);
            lightEffect[0].Parameters["View"].SetValue(rtcView);
            GraphicsDevice.SetRenderTarget(0, renderCube, CubeMapFace.NegativeY);
            GraphicsDevice.Clear(Color.Green);
            DrawModel(walls, Matrix.Identity);
            DrawModel(columns, Matrix.Identity);

            GraphicsDevice.RenderState.CullMode = CullMode.CullCounterClockwiseFace;

            GraphicsDevice.SetRenderTarget(0, null);
            GraphicsDevice.DepthStencilBuffer = oldDS;
            lightEffect[0].Parameters["View"].SetValue(view);
            lightEffect[0].Parameters["Projection"].SetValue(projection);

            envMapping.Parameters["EnvironmentTexture"].SetValue(renderCube.GetTexture());
        }
    }
}
